/*
 * 	What:  Dynamics for 4 player haptic tether ball game
 *           and modal system game for ME161
 *
 *		When: December 1, 1996
 *
 *		Who:	Christopher Richard
 *          Allison Okamura and Weston Griffin 12/2/98
 *
 *		All angles are caluclated in degrees
 *			 thetaX - xangle of pendulum
 *			 dthetaX	- angular velocity in thetax direction
 *	 		 ddthetaX - angular acceleration in thetaX direction
 *
 */

#include <math.h>
#include<stdio.h>
#include<stdlib.h>
#include "dynamics.h"
#include "readcfg.h"

// Defines
#define  XAXIS_ANGLE				90
#define  YAXIS_ANGLE				90
#define  DEG2RAD					(3.1415/180)

/* Config file variables */
float M, R, J, Ct, Fc, G;
float PADDLE_STIFFNESS, PADDLE_DAMPING;
float M1, M2, K1, K2, C1, C2;

/******* variables for tetherDyn **********************/
float fxp, fxn, fyp, fyn;
extern float theta[2];  /* fixme  NUM is in tether.cpp*/
float thetaX = THETAX0, dthetaX, ddthetaX, thetaY = THETAY0, dthetaY, ddthetaY;
float dthetaXold, ddthetaXold, dthetaYold, ddthetaYold;
float thetaPxp, thetaPxn, thetaPyp, thetaPyn;

/********* variables for modalDyn ********************/
float x1[4], dx1[4], ddx1[4], dx1old[4], ddx1old[4];
float x2[4], dx2[4], ddx2[4], dx2old[4], ddx2old[4];
float f[4];
extern float xp[4];

void tetherDyn(float dt, int left, int right)
{
	if (left) {
	/* determine paddle angles */
		  thetaPxp = (-theta[1]/2 + XAXIS_ANGLE);
		  thetaPxn = theta[0]/2 - XAXIS_ANGLE;

	/* determine paddle force */
	if ( thetaX > thetaPxp ) {
		fxp =  PADDLE_STIFFNESS*( thetaX - thetaPxp )+ PADDLE_DAMPING*(dthetaX);
		fxn = 0.0;
	} else if ( thetaX < thetaPxn ) {
		fxn = PADDLE_STIFFNESS*( thetaX - thetaPxn ) + PADDLE_DAMPING*(dthetaX);
		fxp = 0.0;
	} else {
		fxn = fxp = 0.0;
	}

	/* calculate ball acceleration */
	ddthetaXold = ddthetaX;
        ddthetaX = (-Ct*dthetaX - Fc*sign(dthetaX) - fxp*R - fxn*R + M*G*R*thetaX)/J;

	/* calculate ball velocity */
	dthetaXold = dthetaX;
	dthetaX += integrate(ddthetaXold,ddthetaX,dt);

	/* calculate ball position */
	thetaX += integrate(dthetaXold, dthetaX,dt);
	}

	if (right) {

		/* determine paddle angles */
		thetaPyp = -theta[3]/2 + YAXIS_ANGLE;
		thetaPyn	= theta[2]/2 - YAXIS_ANGLE;

		/* determine paddle force */
		if ( thetaY > thetaPyp ) {
			fyp =  PADDLE_STIFFNESS*( thetaY - thetaPyp ) + PADDLE_DAMPING*(dthetaY);
			fyn = 0.0;
		} else if ( thetaY < thetaPyn ) {
			fyn = PADDLE_STIFFNESS*( thetaY - thetaPyn ) + PADDLE_DAMPING*(dthetaY);
			fyp = 0.0;
		} else {
			fyn = fyp = 0.0;
		}

		/* calculate ball acceleration */
      ddthetaYold = ddthetaY;
      ddthetaY = ( -Ct*dthetaY - Fc*sign(dthetaY) - fyp*R - fyn*R + M*G*R*thetaY)/J;

		/* calculate ball velocity */
		dthetaYold = dthetaY;
		dthetaY += integrate(ddthetaYold,ddthetaY,dt);

		/* calculate ball position */
		thetaY += integrate(dthetaYold, dthetaY,dt);
	}
}


float integrate(float lastValue, float currentValue, float dt)
{
	return .5*(lastValue + currentValue)*dt;
}

int sign(float value )
{
	if (value < 0.0 )
		return -1;
	else if (value > 0.0)
		return 1;
	else
		return 0;
}

void modalDyn(float dt)
{
	static float x2old[4];
	int i;

	for (i=0;i<4;i++){

		// store last acceleration
  		ddx1old[i] = ddx1[i];

      // use the position of the paddle as the position of mass 2
      x2[i] = xp[i];

      // calculate the velocity of mass 2
		dx2[i] = (x2[i] - x2old[i])/dt;

		// find the aceleration of mass 1
		ddx1[i] = (K2*(x2[i] - x1[i]) + C2*(dx2[i] - dx1[i]) - K1*x1[i] - C1*dx1[i])/M1;

		// calculate position of mass 1
		dx1old[i] = dx1[i];
		dx1[i] += integrate(ddx1old[i], ddx1[i], dt);
		x1[i] += integrate(dx1old[i], dx1[i], dt);

      // store the old position of mass 2
		x2old[i] = x2[i];

                f[i] = K2*(x1[i] - x2[i]) + C2*(dx1[i] - dx2[i]);

        // old stuff (paddle is atatched to mass 2 through a spring rather
        // than directly attached as above.
      /*
		f[i] = PADDLE_STIFFNESS*(xp[i]-x2[i]);

		ddx1old[i] = ddx1[i];
		ddx2old[i] = ddx2[i];

		ddx1[i] = (K2*(x2[i] -x1[i]) + C2*(dx2[i] - dx1[i]) - K1*x1[i] - C1*dx1[i])/M1;
		ddx2[i] = (f[i] - K2*(x2[i] - x1[i]) - C2*(dx2[i] - dx1[i]))/M2;

		dx1old[i] = dx1[i];
		dx2old[i] = dx2[i];

		dx1[i] += integrate(ddx1old[i], ddx1[i], dt);
		dx2[i] += integrate(ddx2old[i], ddx2[i], dt);

		x1[i] += integrate(dx1old[i], dx1[i], dt);
		x2[i] += integrate(dx2old[i], dx2[i], dt);
      */
	}
}

/*--------------------- initDyn ----------------------------*
 *  Reads constants from data file                          *
 *----------------------------------------------------------*/
void initDyn(void) {

	ReadCfgAddFloat("M",&M,0.5);
	ReadCfgAddFloat("R",&R,1);
	ReadCfgAddFloat("Ct",&Ct,0.5);
	ReadCfgAddFloat("Fc",&Fc,0.1);
	ReadCfgAddFloat("G",&G,0.181);
	ReadCfgAddFloat("PADDLE_STIFFNESS",&PADDLE_STIFFNESS,500);
	ReadCfgAddFloat("PADDLE_DAMPING",&PADDLE_DAMPING,5);
	ReadCfgAddFloat("M1",&M1,2);
	ReadCfgAddFloat("M2",&M2,2);
	ReadCfgAddFloat("K1",&K1,10);
	ReadCfgAddFloat("K2",&K2,10);
        ReadCfgAddFloat("C1",&C1,0.1);
        ReadCfgAddFloat("C2",&C2,0.1);
	ReadCfgExecute("dynamics.dat",RCE_APPLY_DEFAULTS);
	ReadCfgClear();                         // clear config variable list

   // calculate inertia
   J = M*R*R;
}
