/*
 *      modal.cpp
 *
 *      Module to run the ME 161 Haptic Display
 *		  Modal frequency game.
 *
 *      -Christopher Richard and Allison Okamura 12/1/96
 *
 */

/* Includes */
#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include <iostream.h>
#include <math.h>
#include <graphics.h>
#include "display.h"
#include "haptic.h"
#include "ioboard.h"
#include "timing.h"
#include "dynamics.h"
#include "readcfg.h"

/* Defines */
#define NUM       4
#define OFF       0
#define ON        1
#define LO        0
#define HI        1
#define FREQ      10000 // Hz
#define TIMER		0x1C // interrupt for timing
#define WRITE     10   // number of cycles before writing to motors
#define KD        0.001

#define X0        50
#define X10       (X0+150)
#define X20       (X0+300)
#define MSIZE     50
#define RADIUS    10

/* Globals */
void interrupt (*Moldhandler)(...);
void interrupt ModalISR(...);

int MEnabled[NUM] = {OFF, OFF, OFF, OFF};
int MCalibrated[NUM] = {OFF, OFF, OFF, OFF};
float Ma[NUM] = {0.0, 0.0, 0.0, 0.0};
float Mb[NUM] = {0.0, 0.0, 0.0, 0.0};
float Mc[NUM] = {0.0, 0.0, 0.0, 0.0};
float Md[NUM] = {0.0, 0.0, 0.0, 0.0};
int MStation = 0;

float Mfreq;
float xp[NUM], Moutput[NUM];
extern float f[NUM], x1[NUM], x2[NUM];

/*----------------------- ModalISR -------------------------*
 *  Interrupt Service Routine for modal frequency game      *
 *----------------------------------------------------------*/
void interrupt ModalISR(...)
{
    int i, j;
    float sensor[NUM][WRITE], avg_sensor[NUM];
    float angle;
    static float last_xp[NUM] = {0.0, 0.0, 0.0, 0.0};
    static int calc_ctr = 0;
	 float dxp[NUM];

    // Read Sensors
    for (i=0; i<NUM; i++) sensor[i][calc_ctr] = (float) read_ad();

    calc_ctr++;

    if (calc_ctr >= WRITE) {

		 calc_ctr = 0;

		 // for each station
	    for (i=0; i<NUM; i++) {

			// get the average sensor readings
         avg_sensor[i] = 0;
         for (j=0; j<WRITE; j++) avg_sensor[i] += sensor[i][j];
			avg_sensor[i] *= 1.0/WRITE;

        	angle = Ma[i]*avg_sensor[i]*avg_sensor[i]*avg_sensor[i]
               + Mb[i]*avg_sensor[i]*avg_sensor[i]
               + Mc[i]*avg_sensor[i] + Md[i] - 180.0;

                if (MCalibrated[i] == OFF) angle = 0;

			// get position and velocity
         xp[i] = angle*DEG2RAD*LX;
         dxp[i] = (xp[i] - last_xp[i])*Mfreq*WRITE;
		   last_xp[i] = xp[i];
	    }

      // Calculate dynamic interaction
        modalDyn(WRITE/Mfreq);

		for (i=0; i<NUM; i++) {
        //f[i] += -KD*dxp[i];
 			Moutput[i] = f[i]*MEnabled[i]; // calculate output current
     		output_da(i,Moutput[i]);       // send to motor
      }
	 }
}


/*---------------------- ModalOutline ----------------------*
 *  Draws outline for running the modal system game         *
 *----------------------------------------------------------*/
void ModalOutline(void)
{
	 int i;

    setcolor(WHITE);
    outtextxy(10,10,"ME 161 Modal Frequency Game");

    // Instructions
    setcolor(CYAN);
    outtextxy(10,30,"Use 1-4 to select station:");

    outtextxy(10,40,"(c)alibration");
    outtextxy(10,50,"(f)orce enable/disable");
    outtextxy(10,60,"(q)uit");

    // Current Settings
    setcolor(GREEN);
    outtextxy(230,30,"1");
    outtextxy(230,40,"not calibrated");
    outtextxy(230,50,"disabled");

    // Current position
    setcolor(YELLOW);
    outtextxy(400,30,"position:");
    outtextxy(400,40,"output:");
    outtextxy(400,50,"mass 1 pos:");
    outtextxy(400,60,"mass 2 pos:");

    // Draw masses
    setlinestyle(SOLID_LINE,0,NORM_WIDTH);
	 for (i=0; i<NUM; i++) {
		 setcolor(YELLOW);
	    line(X0,(i+1)*110-30,X0,(i+1)*110+30);	// ground
		 setcolor(CYAN);
       rectangle(X10-MSIZE/2,(i+1)*110-MSIZE/2,X10+MSIZE/2,(i+1)*110+MSIZE/2); // mass1
       rectangle(X20-MSIZE/2,(i+1)*110-MSIZE/2,X20+MSIZE/2,(i+1)*110+MSIZE/2); // mass2
  		 setcolor(GREEN);
       line(X0,(i+1)*110-10,X10-MSIZE/2,(i+1)*110-10); // spring 1
       line(X0,(i+1)*110+10,X10-MSIZE/2,(i+1)*110+10); // damper 1
       line(X10+MSIZE/2,(i+1)*110-10,X20-MSIZE/2,(i+1)*110-10); // spring 2
       line(X10+MSIZE/2,(i+1)*110+10,X20-MSIZE/2,(i+1)*110+10); // damper 2
       setcolor(WHITE);
   	 setfillstyle(SOLID_FILL, WHITE);
	    fillellipse(X20+MSIZE/2+RADIUS,(i+1)*110,RADIUS,RADIUS); // finger
	 }
}

/*------------------- DisplayModalGraphics -----------------*
 *  Draws graphics for Modal System game                    *
 *----------------------------------------------------------*/
void DisplayModalGraphics(void)
{
	int i;
   char string[20];
   static float m1_pos[NUM], last_m1_pos[NUM] = {0.0, 0.0, 0.0, 0.0}, last_m1_print = 0.0;
   static float m2_pos[NUM], last_m2_pos[NUM] = {0.0, 0.0, 0.0, 0.0}, last_m2_print = 0.0;
   static float fing_pos[NUM], last_fing_pos[NUM] = {0.0, 0.0, 0.0, 0.0}, last_fing_print = 0.0;
   static float iout[NUM], last_iout = 0.0;

   // make local copy of interrupt variables
   disable();
   for (i=0; i<NUM; i++) {
      m1_pos[i] = x1[i];
      m2_pos[i] = x2[i];
      fing_pos[i] = xp[i];
      iout[i] = Moutput[i];
   }
	enable();

   // erase old data
   setcolor(BLACK);
	sprintf(string,"%3.1f",last_fing_print);
   outtextxy(500,30,string);
	sprintf(string,"%3.1f",last_iout);
   outtextxy(500,40,string);
	sprintf(string,"%3.1f",last_m1_print);
   outtextxy(500,50,string);
	sprintf(string,"%3.1f",last_m2_print);
   outtextxy(500,60,string);

   // draw new data
   setcolor(YELLOW);
	sprintf(string,"%3.1f",fing_pos[MStation]);
 	outtextxy(500,30,string);
	sprintf(string,"%3.1f",iout[MStation]);
  	outtextxy(500,40,string);
	sprintf(string,"%3.1f",m1_pos[MStation]);
  	outtextxy(500,50,string);
	sprintf(string,"%3.1f",m2_pos[MStation]);
  	outtextxy(500,60,string);

   // erase old modal graphics
   for (i=0; i<NUM; i++) {
		setcolor(BLACK);
		rectangle(X10+last_m1_pos[i]-MSIZE/2,(i+1)*110-MSIZE/2,X10+last_m1_pos[i]+MSIZE/2,(i+1)*110+MSIZE/2);
		rectangle(X20+last_m2_pos[i]-MSIZE/2,(i+1)*110-MSIZE/2,X20+last_m2_pos[i]+MSIZE/2,(i+1)*110+MSIZE/2);
		line(X0,(i+1)*110-10,X10+last_m1_pos[i]-MSIZE/2,(i+1)*110-10);
		line(X0,(i+1)*110+10,X10+last_m1_pos[i]-MSIZE/2,(i+1)*110+10);
		line(X10+last_m1_pos[i]+MSIZE/2,(i+1)*110-10,X20+last_m2_pos[i]-MSIZE/2,(i+1)*110-10);
		line(X10+last_m1_pos[i]+MSIZE/2,(i+1)*110+10,X20+last_m2_pos[i]-MSIZE/2,(i+1)*110+10);
   	setfillstyle(SOLID_FILL, BLACK);
		fillellipse(X20+last_fing_pos[i]+MSIZE/2+RADIUS,(i+1)*110,RADIUS,RADIUS); // finger
   }

   // draw new modal graphics
   for (i=0; i<NUM; i++) {
		setcolor(CYAN);
		rectangle(X10+m1_pos[i]-MSIZE/2,(i+1)*110-MSIZE/2,X10+m1_pos[i]+MSIZE/2,(i+1)*110+MSIZE/2);
		rectangle(X20+m2_pos[i]-MSIZE/2,(i+1)*110-MSIZE/2,X20+m2_pos[i]+MSIZE/2,(i+1)*110+MSIZE/2);
		setcolor(GREEN);
		line(X0,(i+1)*110-10,X10+m1_pos[i]-MSIZE/2,(i+1)*110-10);
		line(X0,(i+1)*110+10,X10+m1_pos[i]-MSIZE/2,(i+1)*110+10);
		line(X10+m1_pos[i]+MSIZE/2,(i+1)*110-10,X20+m2_pos[i]-MSIZE/2,(i+1)*110-10);
		line(X10+m1_pos[i]+MSIZE/2,(i+1)*110+10,X20+m2_pos[i]-MSIZE/2,(i+1)*110+10);
      setcolor(WHITE);
		setfillstyle(SOLID_FILL, WHITE);
		fillellipse(X20+fing_pos[i]+MSIZE/2+RADIUS,(i+1)*110,RADIUS,RADIUS); // finger
   }

	// store last positions
   for (i=0; i<NUM; i++) {
      last_m1_pos[i] = m1_pos[i];
      last_m2_pos[i] = m2_pos[i];
      last_fing_pos[i] = fing_pos[i];
   }
   last_m1_print = m1_pos[MStation];
   last_m2_print = m2_pos[MStation];
   last_fing_print = fing_pos[MStation];
   last_iout = iout[MStation];
}

/*------------------------ ReadModalCal --------------------*
 *  Modal System calibration function                       *
 *----------------------------------------------------------*/
void ReadModalCal(int station)
{
	char string1[20], string2[60];
   char astring[3], bstring[3], cstring[3], dstring[3];
   int grp_num;
   int polypoints[8] = {420,140,640,140,640,200,420,200};

   setcolor(CYAN);
   outtextxy(430,150, "Enter the group number");
   sprintf(string1, "for station %d: ", station+1);
   outtextxy(430,160, string1);
	setcolor(WHITE);
	gotoxy(56,12); gets(string2);
  	sscanf(string2, "%i", &grp_num);
  	setcolor(BLACK);
	setlinestyle(SOLID_LINE,0,NORM_WIDTH);
	setfillstyle(SOLID_FILL, BLACK);
	fillpoly(4, polypoints);

	// Read calibration values from data file
	sprintf(astring, "A%i", grp_num);
	sprintf(bstring, "B%i", grp_num);
	sprintf(cstring, "C%i", grp_num);
	sprintf(dstring, "D%i", grp_num);
	ReadCfgAddFloat(astring,&Ma[station],0);
	ReadCfgAddFloat(bstring,&Mb[station],0);
	ReadCfgAddFloat(cstring,&Mc[station],0);
	ReadCfgAddFloat(dstring,&Md[station],0);
	ReadCfgExecute("calib.dat",RCE_APPLY_DEFAULTS);
	ReadCfgClear();				// clear config variable list

   setcolor(BLACK);
   fillpoly(4,polypoints);
}

/*------------------------ ModalSystem ---------------------*
 *  Main Loop for Modal System game                         *
 *----------------------------------------------------------*/
void ModalSystem(void)
{
    char key;
    int done = OFF;
    int i;

    // set the number of channels to scan
    MStation = 0;
    scan_channels(0,(NUM-1));
    for (i=0; i<NUM; i++) MEnabled[i] = OFF;

    // draw initial screen
    init_graphics();
    ModalOutline();

    // read dynamics info
    initDyn();

	 // zero output
    for (i=0; i<NUM; i++) output_da(i,0.0);

    // start interrupt
	 Mfreq = setTimer(FREQ);
	 Moldhandler = getvect(TIMER);
    setvect(TIMER, ModalISR);
	 enable();

    // continuous loop
     while (!done) {

		  // update graphics and printouts
        DisplayModalGraphics();

		  // user input
        if (kbhit()) {
            key = getch();
            switch (key) {
            	 case 'c':
               	  if (MEnabled[MStation]) {
                    	 MEnabled[MStation] = OFF;
                      setcolor(BLACK);
							 outtextxy(230,50,"enabled");
                      setcolor(GREEN);
                   	 outtextxy(230,50,"disabled");
                    }
                    ReadModalCal(MStation);
	                 MCalibrated[MStation] = ON;
						  setcolor(BLACK); outtextxy(230,40,"not calibrated");
   					  setcolor(GREEN); outtextxy(230,40,"calibrated");
                	  break;
                case 'f':
					 	  if (MEnabled[MStation]) {
                    		MEnabled[MStation] = OFF;
							   setcolor(BLACK); outtextxy(230,50,"enabled");
							   setcolor(GREEN); outtextxy(230,50,"disabled");
                    } else if (MCalibrated[MStation]) {
                    		MEnabled[MStation] = ON;
							   setcolor(BLACK); outtextxy(230,50,"disabled");
							   setcolor(GREEN); outtextxy(230,50,"enabled");
                    }
                    break;
					 case '1':
                	  // erase old values
						  setcolor(BLACK);
						  outtextxy(230,30,"2");
						  outtextxy(230,30,"3");
						  outtextxy(230,30,"4");
						  outtextxy(230,40,"calibrated");
						  outtextxy(230,40,"not calibrated");
						  outtextxy(230,50,"enabled");
						  outtextxy(230,50,"disabled");
                    // new values
                    MStation = 0;
                    setcolor(GREEN);
						  outtextxy(230,30,"1");
                    if (MCalibrated[MStation]) outtextxy(230,40,"calibrated");
                    else outtextxy(230,40,"not calibrated");
                    if (MEnabled[MStation]) outtextxy(230,50,"enabled");
                    else outtextxy(230,50,"disabled");
                    break;
					 case '2':
                	  // erase old values
						  setcolor(BLACK);
						  outtextxy(230,30,"1");
						  outtextxy(230,30,"3");
						  outtextxy(230,30,"4");
						  outtextxy(230,40,"calibrated");
						  outtextxy(230,40,"not calibrated");
						  outtextxy(230,50,"enabled");
						  outtextxy(230,50,"disabled");
                    // new values
                    MStation = 1;
                    setcolor(GREEN);
						  outtextxy(230,30,"2");
                    if (MCalibrated[MStation]) outtextxy(230,40,"calibrated");
                    else outtextxy(230,40,"not calibrated");
                    if (MEnabled[MStation]) outtextxy(230,50,"enabled");
                    else outtextxy(230,50,"disabled");
                    break;
					 case '3':
                	  // erase old values
						  setcolor(BLACK);
						  outtextxy(230,30,"1");
						  outtextxy(230,30,"2");
						  outtextxy(230,30,"4");
						  outtextxy(230,40,"calibrated");
						  outtextxy(230,40,"not calibrated");
						  outtextxy(230,50,"enabled");
						  outtextxy(230,50,"disabled");
                    // new values
                    MStation = 2;
                    setcolor(GREEN);
						  outtextxy(230,30,"3");
                    if (MCalibrated[MStation]) outtextxy(230,40,"calibrated");
                    else outtextxy(230,40,"not calibrated");
                    if (MEnabled[MStation]) outtextxy(230,50,"enabled");
                    else outtextxy(230,50,"disabled");
                    break;
					 case '4':
                	  // erase old values
						  setcolor(BLACK);
						  outtextxy(230,30,"1");
						  outtextxy(230,30,"2");
						  outtextxy(230,30,"3");
						  outtextxy(230,40,"calibrated");
						  outtextxy(230,40,"not calibrated");
						  outtextxy(230,50,"enabled");
						  outtextxy(230,50,"disabled");
                    // new values
                    MStation = 3;
                    setcolor(GREEN);
						  outtextxy(230,30,"4");
                    if (MCalibrated[MStation]) outtextxy(230,40,"calibrated");
                    else outtextxy(230,40,"not calibrated");
                    if (MEnabled[MStation]) outtextxy(230,50,"enabled");
                    else outtextxy(230,50,"disabled");
                    break;
                case 'q':
                    done = ON;
                    break;
                default:
                    break;
            }
        }
    }

    // shut down
    getch();
	 disable();
    setvect(TIMER,Moldhandler);
    for (i=0; i<NUM; i++) output_da(i,0.0);
	 closegraph();
}
